unit CSkinnableForm;

interface

uses
  SysUtils, Classes, Forms, Windows, Graphics, Messages;

type
  TCSkinnableForm = class(TComponent)
  private
    FActive:boolean;
    FForm:TForm;

    FTransparent:boolean;

    OldWindowRgn,
    WindowRgn : HRGN;

    acanvas:TCanvas;
    oldwindowproc:TWndMethod;

    FImageFile:string;
    FMaskColorXPos:integer;
    FMaskColorYPos:integer;

    procedure setActive(a:boolean);
    procedure setTransparent(value:boolean);
    procedure setImageFile(value:string);
    procedure setForm(c: TForm);
    function theMaskColor:TColor;
  protected
    OldBorderStyle:TBorderStyle;
    OldborderIcons:TBorderIcons;
    OldHeight:integer;
    OldWidth:integer;
    OldOnpaint:TNotifyEvent;

    ImagePicture: TPicture;
    MaskPicture: TPicture;
    MaskMap  : TBitmap;
    ImageMap : TBitmap;
    function getMaskColor(sl:pbytearray; x:word):TColor;
    procedure doOnPaint;
    procedure restorePaint;
    procedure restoreShape;

    function useImage:boolean;
    function useMask:boolean;

    procedure mywndproc(var msg:TMessage);

    procedure imageChanged(Sender: TObject);
  public
    property Image:TPicture read ImagePicture;
    property Mask:TPicture read MaskPicture;
    procedure doOnCreate;
    constructor create(AOwner:TComponent); override;
    destructor destroy; override;
    procedure refresh(image,mask:string);
    procedure clearImageAndMask;
    procedure setImage(picture:TPicture);
    procedure setMask(picture:TPicture);
    procedure removeImage;
    procedure removeMask;
    procedure doMove(var msg:TMessage); 
  published
    property Form:TForm read FForm write setForm default nil;
    property Active:boolean read FActive write setActive default false;
    property Transparent:boolean read FTransparent write setTransparent;
    property ImageFile:string read FImageFile write setImageFile;
    property MaskColorXPos:integer read FMaskColorXPos write FMaskColorXPos; 
    property MaskColorYPos:integer read FMaskColorYPos write FMaskColorYPos;
    property MaskColor:TColor read theMaskColor; 
  end;

procedure Register;

implementation

uses ImageUtils, Controls, StdCtrls;

procedure Register;
begin
  RegisterComponents('Ciuly''s UI Components', [TCSkinnableForm]);
end;

{ TCSkinnableForm }

procedure TCSkinnableForm.clearImageAndMask;
var wasActive:boolean;
begin
  wasActive:=FActive;
  if wasActive then 
    Active:=false;
  setImage(nil);
  setMask(nil);
//  deleteImages(MaskMap,MaskPicture);
  Active:=wasActive;
end;

constructor TCSkinnableForm.create(AOwner: TComponent);
begin
  inherited;

  ACanvas := TCanvas.Create;
  ImageMap:=nil; // no image
  MaskMap:=nil; // no mask
  ImagePicture:=nil;
  MaskPicture:=nil;
  FTransparent:=false;
  OldHeight:=0;
  OldWidth:=0;
  if Active then
    dooncreate;

  if not (AOwner is TForm) then
  begin
    destroy;
    raise Exception.Create('The owner of TCSkinableForm MUST be a TForm or descendant!');
  end;
  Form:=TForm(AOwner);

  //to Show window contents while dragging:
  SystemParametersInfo(SPI_SETDRAGFULLWINDOWS, 1, nil, 0);

  //to disable this option call the function:
  //SystemParametersInfo(SPI_SETDRAGFULLWINDOWS, 0, nil, 0);
end;

destructor TCSkinnableForm.destroy;
begin
  try
    setForm(nil);
  except
  end;
  deleteImages(ImageMap,ImagePicture);
  deleteImages(MaskMap,MaskPicture);
  ACanvas.free;
  inherited;
end;

procedure TCSkinnableForm.doMove(var msg: TMessage);
begin
  msg.result:=HTCAPTION;
end;

procedure TCSkinnableForm.doOnCreate;

  {search scanline for a run of like pixels. Really, just gets colour
  of pixel at spanleft, then runs along the scanline until it finds
  a pixel which isn't the same colour.

  sl - Pointer to TScanline data (I think)
  spanleft - pixel count index into scanline of first pixel in span.
  spanright - pixel count index into scanline of last pixel in span.
              Determined and returned by this routine.
  spancolor - color of pixels in the span. Determined and returned
              by this routine.
  }
  procedure GetNextSpan(const sl: PByteArray; const spanleft:integer; var spanright:integer; var spancolor:tcolor);
  var
      column : integer;
      TestColor : TColor;
  begin
    with FForm do
    begin
      column := spanleft;
      TestColor := getMaskcolor(sl,spanleft);
      repeat
        column := column + 1;
      until (testColor<>getMaskColor(sl,column)) or (column>=width);
      spanright := column - 1;
      spancolor := TestColor;
    end;
  end;

var
    RowRgn    : HRGN;
    SpanRgn   : HRGN;
    row : integer;
    sl : PByteArray;
    spanleft  : integer;
    spanright : integer;
    spancolor : tcolor;
begin
  if useMask and Active then
    with FForm do
    begin
      BorderStyle := bsNone;
      BorderIcons := [];

      {Set window size to fit mask bitmap}
      if ImageMap<>nil then
      begin
        height := ImageMap.Height;
        width  := ImageMap.width;
      end;

      {create empty region for window. Create empty region for scanline.
      Search for spans of opaque pixels. Create a region for a span. Add
      it to the scanline region. When scanline done, add it to the window.
      Move to next scanline.}
      WindowRgn := CreateRectRgn(0,0,0,0);
      for row := 0 to MaskMap.Height-1 do
        begin
          RowRgn := CreateRectRgn(0,row,0,row);
          sl := maskMap.scanline[row];
          spanleft := 0;
          spanright := 0;

          repeat
            GetNextSpan(sl,spanleft,spanright,spancolor);
            if spancolor<>MaskColor then
            begin
              SpanRgn := CreateRectRgn(spanleft,row,spanright+1,row+1);
              CombineRgn(RowRgn, RowRgn, SpanRgn, RGN_OR);
              DeleteObject(SpanRgn);
            end;
            spanleft:=spanright+1;
          until (spanright >= width);

          CombineRgn(WindowRgn, WindowRgn, RowRgn, RGN_OR);
          {Delete all created regions except the final window region}
          DeleteObject(RowRgn);
        end; {row looping}
      GetWindowRgn(Handle,OldWindowRgn);
      SetWindowRgn(Handle,WindowRgn,true);
    end;
  doOnPaint;
end;

procedure TCSkinnableForm.doOnPaint;
var i:integer;
begin
  if not useImage then
    exit;
  try
    {to retrieve the device context for the Form1 window}
    ACanvas.Handle := GetWindowDC(FForm.Handle);
    Acanvas.Draw(0,0,ImageMap);
  finally
    {to release the device context}
    ReleaseDC(FForm.Handle, ACanvas.Handle);
  end;
  for i:=1 to FForm.ControlCount do
    FForm.Controls[i-1].Repaint;
end;

procedure TCSkinnableForm.imageChanged(Sender: TObject);
begin
  FForm.Tag:=1;
  FForm.Repaint;
end;

{Returns the colour of a pixel at a given x position on a scanline.
This is the code that assumes three bytes per pixel of rgb data.}
function TCSkinnableForm.getMaskColor(sl: pbytearray; x: word): TColor;
var
  r,g,b:byte;
begin
  result := MaskColor;
  if maskmap = nil then exit;
  if x > MaskMap.width then exit;

  {Pixels are three bytes wide. Move x to locate actual data. Add
  proper offsets for r, g and b.}
  b := sl[x * 3];
  g := sl[x * 3 + 1];
  r := sl[x * 3 + 2];

  {Values used were determined by experimentation on a sample data
  set of 1 image. They reflect a desire to get most fairly pale,
  brigh colours as white; most fairly bright not too poorly
  saturated colours as whatever colour they represent. Note that
  it seems like a low saturation bright cyan will qualify as red.
  Presume this means the code assumes that only pure primaries
  will be encountered. You don't like the tests, write your own.}
{  if (r>=148) and (g>=148) and (b>=148) then result := clWhite
  else if (r>=148) and (g<=200) and (b<=200) then result := clRed
  else if (r<=108) and (g<=108) and (b>=148) then result := clblue
  else if (r<=108) and (g>=148) and (b<=108) then result := clgreen
  else if (r<=108) and (g<=108) and (b<=108) then result := clBlack;{}
  result:=RGB(r,g,b);
end;

procedure TCSkinnableForm.mywndproc(var msg: TMessage);
const max=24;
      messages:array[1..max] of cardinal=( WM_PAINT, WM_NCPAINT, CM_FocusChanged, CN_CtlColorBtn, WM_KILLFOCUS,
                                       WM_SETFOCUS, WM_CTLCOLORBTN, CM_ACTIVATE, CM_DEACTIVATE, CM_ACTIVATE,
                                       CM_DEACTIVATE, CM_GOTFOCUS, CM_LOSTFOCUS, CM_FOCUSCHANGED, CM_VISIBLECHANGED,
                                       CM_ENABLEDCHANGED, CM_COLORCHANGED, CM_CTL3DCHANGED,
                                       WM_ACTIVATE, WM_ENABLE, WM_NCACTIVATE, WM_SETTEXT, WM_ERASEBKGND,
                                       WM_CTLCOLOR);
//  BN_PAINT BN_HILITE BN_UNHILITE BN_DISABLE BN_PUSHED BN_UNPUSHED BN_SETFOCUS BN_KILLFOCUS BM_SETSTATE BM_SETSTYLE BM_SETIMAGE BM_SETCHECK
  function isToPaint:boolean;
  var i:integer;
  begin
    i:=1;
    while (i<=max) and (msg.Msg<>messages[i]) do
      inc(i);                                   
    result:=(i<=max) or ({msg.msg=WM_LBUTTONUP{}(msg.msg>=BM_GETCHECK) and (msg.Msg<=BM_SETIMAGE){});
  end;

  procedure doWndProc;
  var i:integer; l:TLabel; f:TFont; c:TCanvas; b:boolean;
  begin
    if (csDesigning in ComponentState) or (not active) or (not useImage) then
    begin
      oldwindowproc(msg);
      exit;
    end;{}

    b:=isToPaint;
    if b and (tag>0) then
    begin
      tag:=tag-1;
      exit;
    end; // ignore this paint message

    oldwindowproc(msg);
    if b then
    begin
      with FForm do
      begin
        c:=canvas;
        canvas.Draw(0,0,ImageMap);
      end;
      for i:=1 to FForm.ControlCount do
        if (FForm.Controls[i-1] is TLabel) and (FForm.Controls[i-1].Visible) then
        begin
          if c=nil then continue;
          l:=FForm.Controls[i-1] as TLabel;
          f:=c.Font;
          c.Font:=l.Font;
          if l.Transparent then
            c.Brush.Style:=bsClear;
          c.TextOut(l.Left,l.Top,l.Caption);
          c.Font:=f;
        end;
    end;
  end;

begin
  if (msg.Msg=WM_NCHITTEST) and (not (csDesigning in ComponentState)) then
  begin
    if Active then doMove(msg)
              else doWndProc;
  end                                                                 else
    doWndProc;
end;

procedure TCSkinnableForm.refresh(image, mask: string);
var wasActive:boolean; 
begin
  wasActive:=FActive;
  Active:=false;
  if image<>'' then
  begin
    deleteImages(ImageMap,ImagePicture);
    setImage(getPicture(image));
  end;
  if mask<>'' then
  begin
    deleteImages(MaskMap,MaskPicture);
    setMask(getPicture(mask));
  end;
  Active:=wasActive;
end;

procedure TCSkinnableForm.removeImage;
begin
  deleteImages(ImageMap,ImagePicture);
  if MaskMap=nil then
    Active:=false;
end;

procedure TCSkinnableForm.removeMask;
begin
  if ImageMap=nil then
    Active:=false;
  deleteImages(MaskMap,MaskPicture);
end;

procedure TCSkinnableForm.restorePaint;
begin
  FForm.WindowProc:=oldwindowproc;
end;

procedure TCSkinnableForm.restoreShape;
begin
  SetWindowRgn(FForm.Handle,OldWindowRgn,true);
  with FForm do
  begin
    BorderStyle := OldBorderStyle;
    BorderIcons := OldBorderIcons;
    if OldHeight<>0 then
    begin
      Height:=OldHeight;
      Width:=OldWidth;
    end;
  end;
end;

procedure TCSkinnableForm.setActive(a: boolean);
begin
  if csDesigning in ComponentState then
    exit;
  if FActive=a then
    exit;
  FActive:=a and (FForm<>nil);
  if Factive then
  begin
    if FForm<>nil then
    begin
      oldwindowproc:=FForm.WindowProc;
      FForm.WindowProc:=mywndproc;
    end;
    doOnCreate;
  end        else
  begin
    restorePaint;
    restoreShape;
  end;
end;

procedure TCSkinnableForm.setForm(c: TForm);
begin
  if FForm<>nil then
    exit; // the form must not be modified by the user: it must remain the TForm on wich the component was dropped
  if c=FForm then exit;
  if active then
    active:=false;
  if c=nil then
  begin
    if FForm<>nil then
      FForm.Tag:=0;
    OldHeight:=0;
    OldWidth:=0;
  end      else
  begin
    with c do
    begin
      Tag:=integer(self);

      OldHeight:=Height;
      OldWidth:=Width;

      OldBorderStyle := BorderStyle;
      OldBorderIcons := BorderIcons;
    end;
  end;
  FForm:=c;
end;

procedure TCSkinnableForm.setImage(picture: TPicture);
begin
  if picture<>ImagePicture then
  begin
    deleteImages(ImageMap,ImagePicture);
    ImagePicture:=picture;
    if picture<>nil then
    begin
      ImagePicture.Graphic.Transparent:=Transparent;
      ImageMap:=getBitmap(ImagePicture);
      ImageMap.OnChange:=imageChanged;
      with FForm do
      begin
        height := ImageMap.Height;
        width  := ImageMap.width;
      end;
    end             else
      with FForm do
      begin
        height := OldHeight;
        width  := OldWidth;
    end;
    FForm.Repaint;
  end;
end;

procedure TCSkinnableForm.setImageFile(value: string);
begin
  FImageFile:=value;
end;

procedure TCSkinnableForm.setMask(picture: TPicture);
begin
  if picture<>MaskPicture then
  begin
    restoreShape;
    deleteImages(MaskMap,MaskPicture);
    MaskPicture:=picture;
    if picture<>nil then
    begin
      MaskMap:=getBitmap(MaskPicture);
      doOnCreate;
    end;
    if ImageMap<>nil then
      with FForm do
      begin
        height := ImageMap.Height;
        width  := ImageMap.width;
      end;
    SendMessage(FForm.Handle,WM_PAINT,0,0);
  end;
end;

procedure TCSkinnableForm.setTransparent(value: boolean);
begin
  if FTransparent <> value then
    FTransparent:=value;
end;

function TCSkinnableForm.useImage: boolean;
begin
  result:=ImageMap<>nil;
end;

function TCSkinnableForm.useMask: boolean;
begin
  result:=MaskMap<>nil;
end;

function TCSkinnableForm.theMaskColor: TColor;
begin
  if MaskMap<>nil then result:=MaskMap.Canvas.Pixels[FMaskColorXPos, FMaskColorYPos]
                  else result:=clBlack;
end;

end.
 